-- File generated by the BNF Converter (bnfc 2.9.5).

{-# LANGUAGE CPP #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE LambdaCase #-}
#if __GLASGOW_HASKELL__ <= 708
{-# LANGUAGE OverlappingInstances #-}
#endif

-- | Pretty-printer for Sit.

module Sit.Print where

import Prelude
  ( ($), (.)
  , Bool(..), (==), (<)
  , Int, Integer, Double, (+), (-), (*)
  , String, (++)
  , ShowS, showChar, showString
  , all, elem, foldr, id, map, null, replicate, shows, span
  )
import Data.Char ( Char, isSpace )
import qualified Sit.Abs

-- | The top-level printing method.

printTree :: Print a => a -> String
printTree = render . prt 0

type Doc = [ShowS] -> [ShowS]

doc :: ShowS -> Doc
doc = (:)

render :: Doc -> String
render d = rend 0 False (map ($ "") $ d []) ""
  where
  rend
    :: Int        -- ^ Indentation level.
    -> Bool       -- ^ Pending indentation to be output before next character?
    -> [String]
    -> ShowS
  rend i p = \case
      "["      :ts -> char '[' . rend i False ts
      "("      :ts -> char '(' . rend i False ts
      "{"      :ts -> onNewLine i     p . showChar   '{'  . new (i+1) ts
      "}" : ";":ts -> onNewLine (i-1) p . showString "};" . new (i-1) ts
      "}"      :ts -> onNewLine (i-1) p . showChar   '}'  . new (i-1) ts
      [";"]        -> char ';'
      ";"      :ts -> char ';' . new i ts
      t  : ts@(s:_) | closingOrPunctuation s
                   -> pending . showString t . rend i False ts
      t        :ts -> pending . space t      . rend i False ts
      []           -> id
    where
    -- Output character after pending indentation.
    char :: Char -> ShowS
    char c = pending . showChar c

    -- Output pending indentation.
    pending :: ShowS
    pending = if p then indent i else id

  -- Indentation (spaces) for given indentation level.
  indent :: Int -> ShowS
  indent i = replicateS (2*i) (showChar ' ')

  -- Continue rendering in new line with new indentation.
  new :: Int -> [String] -> ShowS
  new j ts = showChar '\n' . rend j True ts

  -- Make sure we are on a fresh line.
  onNewLine :: Int -> Bool -> ShowS
  onNewLine i p = (if p then id else showChar '\n') . indent i

  -- Separate given string from following text by a space (if needed).
  space :: String -> ShowS
  space t s =
    case (all isSpace t', null spc, null rest) of
      (True , _   , True ) -> []              -- remove trailing space
      (False, _   , True ) -> t'              -- remove trailing space
      (False, True, False) -> t' ++ ' ' : s   -- add space if none
      _                    -> t' ++ s
    where
      t'          = showString t []
      (spc, rest) = span isSpace s

  closingOrPunctuation :: String -> Bool
  closingOrPunctuation [c] = c `elem` closerOrPunct
  closingOrPunctuation _   = False

  closerOrPunct :: String
  closerOrPunct = ")],;"

parenth :: Doc -> Doc
parenth ss = doc (showChar '(') . ss . doc (showChar ')')

concatS :: [ShowS] -> ShowS
concatS = foldr (.) id

concatD :: [Doc] -> Doc
concatD = foldr (.) id

replicateS :: Int -> ShowS -> ShowS
replicateS n f = concatS (replicate n f)

-- | The printer class does the job.

class Print a where
  prt :: Int -> a -> Doc

instance {-# OVERLAPPABLE #-} Print a => Print [a] where
  prt i = concatD . map (prt i)

instance Print Char where
  prt _ c = doc (showChar '\'' . mkEsc '\'' c . showChar '\'')

instance Print String where
  prt _ = printString

printString :: String -> Doc
printString s = doc (showChar '"' . concatS (map (mkEsc '"') s) . showChar '"')

mkEsc :: Char -> Char -> ShowS
mkEsc q = \case
  s | s == q -> showChar '\\' . showChar s
  '\\' -> showString "\\\\"
  '\n' -> showString "\\n"
  '\t' -> showString "\\t"
  s -> showChar s

prPrec :: Int -> Int -> Doc -> Doc
prPrec i j = if j < i then parenth else id

instance Print Integer where
  prt _ x = doc (shows x)

instance Print Double where
  prt _ x = doc (shows x)

instance Print Sit.Abs.Ident where
  prt _ (Sit.Abs.Ident i) = doc $ showString i
instance Print Sit.Abs.Prg where
  prt i = \case
    Sit.Abs.Prg decls -> prPrec i 0 (concatD [prt 0 decls])

instance Print Sit.Abs.Decl where
  prt i = \case
    Sit.Abs.Sig id_ exp -> prPrec i 0 (concatD [prt 0 id_, doc (showString ":"), prt 0 exp])
    Sit.Abs.Def id_ exp -> prPrec i 0 (concatD [prt 0 id_, doc (showString "="), prt 0 exp])
    Sit.Abs.Open qualid -> prPrec i 0 (concatD [doc (showString "open"), doc (showString "import"), prt 0 qualid])
    Sit.Abs.Blank -> prPrec i 0 (concatD [])

instance Print Sit.Abs.QualId where
  prt i = \case
    Sit.Abs.Sg id_ -> prPrec i 0 (concatD [prt 0 id_])
    Sit.Abs.Cons qualid id_ -> prPrec i 0 (concatD [prt 0 qualid, doc (showString "."), prt 0 id_])

instance Print [Sit.Abs.Decl] where
  prt _ [] = concatD []
  prt _ [x] = concatD [prt 0 x]
  prt _ (x:xs) = concatD [prt 0 x, doc (showString "--;"), prt 0 xs]

instance Print Sit.Abs.IdU where
  prt i = \case
    Sit.Abs.Id id_ -> prPrec i 0 (concatD [prt 0 id_])
    Sit.Abs.Under -> prPrec i 0 (concatD [doc (showString "_")])

instance Print Sit.Abs.Bind where
  prt i = \case
    Sit.Abs.BIrrel id_ -> prPrec i 0 (concatD [doc (showString "."), prt 0 id_])
    Sit.Abs.BRel id_ -> prPrec i 0 (concatD [doc (showString ".."), prt 0 id_])
    Sit.Abs.BAnn ids exp -> prPrec i 0 (concatD [doc (showString "("), prt 0 ids, doc (showString ":"), prt 0 exp, doc (showString ")")])

instance Print [Sit.Abs.Bind] where
  prt _ [] = concatD []
  prt _ [x] = concatD [prt 0 x]
  prt _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Sit.Abs.Ident] where
  prt _ [] = concatD []
  prt _ [x] = concatD [prt 0 x]
  prt _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Sit.Abs.IdU] where
  prt _ [] = concatD []
  prt _ [x] = concatD [prt 0 x]
  prt _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print Sit.Abs.Exp where
  prt i = \case
    Sit.Abs.Var idu -> prPrec i 2 (concatD [prt 0 idu])
    Sit.Abs.Int n -> prPrec i 2 (concatD [prt 0 n])
    Sit.Abs.Infty -> prPrec i 2 (concatD [doc (showString "oo")])
    Sit.Abs.Nat -> prPrec i 2 (concatD [doc (showString "Nat")])
    Sit.Abs.Set -> prPrec i 2 (concatD [doc (showString "Set")])
    Sit.Abs.Set1 -> prPrec i 2 (concatD [doc (showString "Set1")])
    Sit.Abs.Set2 -> prPrec i 2 (concatD [doc (showString "Set2")])
    Sit.Abs.Zero -> prPrec i 2 (concatD [doc (showString "zero")])
    Sit.Abs.Suc -> prPrec i 2 (concatD [doc (showString "suc")])
    Sit.Abs.Fix -> prPrec i 2 (concatD [doc (showString "fix")])
    Sit.Abs.LZero -> prPrec i 2 (concatD [doc (showString "lzero")])
    Sit.Abs.LSuc -> prPrec i 2 (concatD [doc (showString "lsuc")])
    Sit.Abs.Size -> prPrec i 0 (concatD [doc (showString "Size")])
    Sit.Abs.App exp1 exp2 -> prPrec i 1 (concatD [prt 1 exp1, prt 2 exp2])
    Sit.Abs.Lam idus exp -> prPrec i 0 (concatD [doc (showString "\\"), prt 0 idus, doc (showString "->"), prt 0 exp])
    Sit.Abs.Forall binds exp -> prPrec i 0 (concatD [doc (showString "forall"), prt 0 binds, doc (showString "->"), prt 0 exp])
    Sit.Abs.Pi exp1 exp2 exp3 -> prPrec i 0 (concatD [doc (showString "("), prt 0 exp1, doc (showString ":"), prt 0 exp2, doc (showString ")"), doc (showString "->"), prt 0 exp3])
    Sit.Abs.Arrow exp1 exp2 -> prPrec i 0 (concatD [prt 1 exp1, doc (showString "->"), prt 0 exp2])
    Sit.Abs.Case exp1 exp2 exp3 -> prPrec i 0 (concatD [doc (showString "case"), prt 0 exp1, doc (showString "return"), prt 0 exp2, doc (showString "of"), prt 0 exp3])
    Sit.Abs.Plus exp n -> prPrec i 0 (concatD [prt 1 exp, doc (showString "+"), prt 0 n])
    Sit.Abs.ELam exp1 idu exp2 -> prPrec i 0 (concatD [doc (showString "\\"), doc (showString "{"), doc (showString "("), doc (showString "zero"), doc (showString "_"), doc (showString ")"), doc (showString "->"), prt 0 exp1, doc (showString ";"), doc (showString "("), doc (showString "suc"), doc (showString "_"), prt 0 idu, doc (showString ")"), doc (showString "->"), prt 0 exp2, doc (showString "}")])
